@@ -140,7 +140,8 @@ template<typename T> T* DeclarationBuilder::eventuallyReopenDeclaration(Identifi
     return static_cast<T*>(dec);
 }
 
-template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Ast* node, Declaration* previous, AbstractType::Ptr type)
+template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Ast* node, Declaration* previous,
+                                                                     AbstractType::Ptr type, VisitVariableFlags flags)
 {
     if ( node->astType == Ast::NameAstType ) {
         NameAst* currentVariableDefinition = static_cast<NameAst*>(node);
@@ -154,10 +155,10 @@ template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Ast* node,
             return 0;
         }
         Identifier* id = currentVariableDefinition->identifier;
-        return visitVariableDeclaration<T>(id, currentVariableDefinition, previous, type);
+        return visitVariableDeclaration<T>(id, currentVariableDefinition, previous, type, flags);
     }
     else if ( node->astType == Ast::IdentifierAstType ) {
-        return visitVariableDeclaration<T>(static_cast<Identifier*>(node), 0, previous, type);
+        return visitVariableDeclaration<T>(static_cast<Identifier*>(node), nullptr, previous, type, flags);
     }
     else {
         qCWarning(KDEV_PYTHON_DUCHAIN) << "cannot create variable declaration for non-(name|identifier) AST, this is a programming error";
@@ -166,12 +167,12 @@ template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Ast* node,
 }
 
 template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Identifier* node, RangeInRevision range,
-                                                                     AbstractType::Ptr type)
+                                                                     AbstractType::Ptr type, VisitVariableFlags flags)
 {
     Ast pseudo;
     pseudo.startLine = range.start.line; pseudo.startCol = range.start.column;
     pseudo.endLine = range.end.line; pseudo.endCol = range.end.column;
-    T* result = visitVariableDeclaration<T>(node, &pseudo, 0, type);
+    T* result = visitVariableDeclaration<T>(node, &pseudo, nullptr, type, flags);
     return result;
 }
 
@@ -249,7 +250,8 @@ template<typename T> QList<Declaration*> DeclarationBuilder::reopenFittingDeclar
 }
 
 typedef QPair<Declaration*, int> p;
-template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Identifier* node, Ast* originalAst, Declaration* previous, AbstractType::Ptr type)
+template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Identifier* node, Ast* originalAst, Declaration* previous,
+                                                                     AbstractType::Ptr type, VisitVariableFlags flags)
 {
     DUChainWriteLocker lock;
     Ast* rangeNode = originalAst ? originalAst : node;
@@ -278,6 +280,9 @@ template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Identifier*
     Declaration* dec = 0;
     existingDeclarations = reopenFittingDeclaration<T>(existingDeclarations, kindForType(type), range, &dec);
     bool declarationOpened = (bool) dec;
+    if ( flags & AbortIfReopenMismatch && previous && ! declarationOpened ) {
+        return nullptr;
+    }
     
     // tells whether the declaration found for updating is in the same top context
     bool inSameTopContext = true;
@@ -321,7 +326,7 @@ template<typename T> T* DeclarationBuilder::visitVariableDeclaration(Identifier*
         AbstractType::Ptr newType;
         if ( currentContext()->type() == DUContext::Function ) {
             // check for argument type hints (those are created when calling functions)
-            AbstractType::Ptr hints = Helper::extractTypeHints(dec->abstractType(), topContext());
+            AbstractType::Ptr hints = Helper::extractTypeHints(dec->abstractType());
             qCDebug(KDEV_PYTHON_DUCHAIN) << hints->toString();
             if ( hints.cast<IndexedContainer>() || hints.cast<ListType>() ) {
                 // This only happens when the type hint is a tuple, which means the vararg/kwarg of a function is being processed.
@@ -554,6 +559,7 @@ void DeclarationBuilder::visitComprehension(ComprehensionAst* node)
     // TODO add a special case to the usebuilder to display the second occurence as a declaration
     RangeInRevision declarationRange(currentContext()->range().start, currentContext()->range().start);
     declarationRange.end.column -= 1;
+    declarationRange.start.column -= 1;
     
     AbstractType::Ptr targetType(new IntegralType(IntegralType::TypeMixed));
     if ( node->iterator ) {
@@ -1193,93 +1199,32 @@ QList<ExpressionAst*> DeclarationBuilder::targetsOfAssignment(QList<ExpressionAs
     return lhsExpressions;
 }
 
-QList<DeclarationBuilder::SourceType> DeclarationBuilder::sourcesOfAssignment(ExpressionAst* items,
-                                                                              int fillWhenLengthMissing) const
+QVector<DeclarationBuilder::SourceType>
+    DeclarationBuilder::unpackAssignmentSource(const SourceType& source,
+                                               int fillWhenLengthMissing) const
 {
-    QList<SourceType> sources;
-    QList<ExpressionAst*> values;
-
-    if ( items && items->astType == Ast::TupleAstType ) {
-        values = static_cast<TupleAst*>(items)->elements;
-    }
-    else {
-        // This handles the a, b, c = [1, 2, 3] case. Since the assignment can also be like
-        // d = [1, 2, 3]; a, b, c = d we can't generally know the amount of elements
-        // in the right operand; so all elements are treated to have the same type.
-        if ( fillWhenLengthMissing > 0 ) {
-            ExpressionVisitor v(currentContext());
-            v.visitNode(items);
-            auto container = ListType::Ptr::dynamicCast(v.lastType());
-            if ( container ) {
-                AbstractType::Ptr content = container->contentType().abstractType();
-                for ( ; fillWhenLengthMissing != 0; fillWhenLengthMissing-- ) {
-                    sources << SourceType{ content, KDevelop::DeclarationPointer(), false };
-                }
-                return sources;
-            }
+    QVector<SourceType> sources;
+    if ( const IndexedContainer::Ptr container = source.type.cast<IndexedContainer>() ) {
+        // RHS is a tuple or similar, unpack that.
+        for (int i = 0; i < container->typesCount(); ++i) {
+            sources << SourceType{
+                container->typeAt(i).abstractType(),
+                DeclarationPointer(),
+                false
+            };
         }
-
-        // Otherwise, proceed normally.
-        values << items;
     }
-
-    foreach ( ExpressionAst* value, values ) {
-        ExpressionVisitor v(currentContext());
-        v.visitNode(value);
-
-        sources << SourceType{
-            v.lastType(),
-            DeclarationPointer(Helper::resolveAliasDeclaration(v.lastDeclaration().data())),
-            v.isAlias()
-        };
+    else if ( auto container = ListType::Ptr::dynamicCast(source.type) ) {
+        // RHS is a list or similar, can't tell contents apart.
+        // Return content type * requested length.
+        AbstractType::Ptr content = container->contentType().abstractType();
+        for ( ; fillWhenLengthMissing != 0; fillWhenLengthMissing-- ) {
+            sources << SourceType{ content, DeclarationPointer(), false };
+        }
     }
     return sources;
 }
 
-DeclarationBuilder::SourceType DeclarationBuilder::selectSource(const QList<ExpressionAst*>& targets,
-                                                                const QList<DeclarationBuilder::SourceType>& sources,
-                                                                int index, ExpressionAst* rhs) const
-{
-    bool canUnpack = targets.length() == sources.length();
-    SourceType element;
-    // If the length of the right and the left side matches, exact unpacking can be done.
-    // example code: a, b, c = 3, 4, 5
-    // If the left side only contains one entry, unpacking never happens, and the left side
-    // is instead assigned a container type if applicable
-    // example code: a = 3, 4, 5
-    if ( canUnpack ) {
-        element = sources.at(index);
-    }
-    else if ( targets.length() == 1 ) {
-        ExpressionVisitor v(currentContext());
-        v.visitNode(rhs);
-        element = SourceType{
-            v.lastType(),
-            DeclarationPointer(Helper::resolveAliasDeclaration(v.lastDeclaration().data())),
-            v.isAlias()
-        };
-    }
-    else if ( ! sources.isEmpty() ) {
-        // the assignment is of the form "foo, bar, ... = ..." (tuple unpacking)
-        // this one is for the case that the tuple unpacking is not written down explicitly, for example
-        // a = (1, 2, 3); b, c, d = a
-        // the other case (b, c, d = 1, 2, 3) is handled above.
-        if ( const IndexedContainer::Ptr container = sources.first().type.cast<IndexedContainer>() ) {
-            if ( container->typesCount() == targets.length() ) {
-                element.type = container->typeAt(index).abstractType();
-                element.isAlias = false;
-            }
-        }
-    }
-    if ( ! element.type ) {
-        // use mixed if none of the previous ways of determining the type worked.
-        element.type = AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed));
-        element.declaration = nullptr;
-        element.isAlias = false;
-    }
-    return element;
-}
-
 void DeclarationBuilder::assignToName(NameAst* target, const DeclarationBuilder::SourceType& element)
 {
     if ( element.isAlias ) {
@@ -1292,6 +1237,10 @@ void DeclarationBuilder::assignToName(NameAst* target, const DeclarationBuilder:
     else {
         DUChainWriteLocker lock;
         Declaration* dec = visitVariableDeclaration<Declaration>(target, 0, element.type);
+        if ( dec && m_lastComment && ! m_lastComment->usedAsComment ) {
+            dec->setComment(m_lastComment->value);
+            m_lastComment->usedAsComment = true;
+        }
         /** DEBUG **/
         if ( element.type && dec ) {
             Q_ASSERT(dec->abstractType());
@@ -1332,25 +1281,12 @@ void DeclarationBuilder::assignToSubscript(SubscriptAst* subscript, const Declar
 
 void DeclarationBuilder::assignToAttribute(AttributeAst* attrib, const DeclarationBuilder::SourceType& element)
 {
-    // check whether the current attribute is undeclared, but the previos ones known
-    // like in X.Y.Z = 3 where X and Y are defined, but Z isn't; then declare Z.
-    ExpressionVisitor checkForUnknownAttribute(currentContext());
-    checkForUnknownAttribute.visitNode(attrib);
-    Declaration* haveDeclaration = 0;
-    {
-        DUChainReadLocker lock;
-        haveDeclaration = checkForUnknownAttribute.lastDeclaration().data();
-    }
-
-    // declare the attribute.
-    // however, if there's an earlier declaration which does not match the current position
-    // (so it's really a different declaration) we skip this.
+    // visit the base expression before the dot
     ExpressionVisitor checkPreviousAttributes(currentContext());
-    // go "down one level", so only visit "X.Y"
     checkPreviousAttributes.visitNode(attrib->value);
+    DeclarationPointer parentObjectDeclaration = checkPreviousAttributes.lastDeclaration();
 
     DUContextPointer internal(0);
-    DeclarationPointer parentObjectDeclaration = checkPreviousAttributes.lastDeclaration();
 
     if ( ! parentObjectDeclaration ) {
         qCDebug(KDEV_PYTHON_DUCHAIN) << "No declaration for attribute base, aborting creation of attribute";
@@ -1362,6 +1298,7 @@ void DeclarationBuilder::assignToAttribute(AttributeAst* attrib, const Declarati
     }
     // while this is like A = foo(); A.bar = 3
     else {
+        DUChainReadLocker lock;
         StructureType::Ptr structure(parentObjectDeclaration->abstractType().cast<StructureType>());
         if ( ! structure || ! structure->declaration(topContext()) ) {
             return;
@@ -1374,61 +1311,123 @@ void DeclarationBuilder::assignToAttribute(AttributeAst* attrib, const Declarati
         return;
     }
 
-    DUContext* previousContext = currentContext();
-    bool isAlreadyOpen = contextAlreayOpen(internal);
-    if ( isAlreadyOpen ) {
-        activateAlreadyOpenedContext(internal);
-        visitVariableDeclaration<ClassMemberDeclaration>(
-            attrib->attribute, attrib, haveDeclaration, element.type
-        );
-        closeAlreadyOpenedContext(internal);
+    Declaration* attributeDeclaration = nullptr;
+    {
+        DUChainReadLocker lock;
+        attributeDeclaration = Helper::accessAttribute(parentObjectDeclaration.data(),
+                                                       attrib->attribute->value, currentContext());
+    }
+
+    if ( ! attributeDeclaration || ! wasEncountered(attributeDeclaration) ) {
+        // inject a new attribute into the class type
+        DUContext* previousContext = currentContext();
+        bool isAlreadyOpen = contextAlreayOpen(internal);
+        if ( isAlreadyOpen ) {
+            activateAlreadyOpenedContext(internal);
+            visitVariableDeclaration<ClassMemberDeclaration>(
+                attrib->attribute, attrib, attributeDeclaration, element.type, AbortIfReopenMismatch
+            );
+            closeAlreadyOpenedContext(internal);
+        }
+        else {
+            injectContext(internal.data());
+
+            Declaration* dec = visitVariableDeclaration<ClassMemberDeclaration>(
+                attrib->attribute, attrib, attributeDeclaration, element.type, AbortIfReopenMismatch
+            );
+            if ( dec ) {
+                dec->setRange(RangeInRevision(internal->range().start, internal->range().start));
+                dec->setAutoDeclaration(true);
+                DUChainWriteLocker lock;
+                previousContext->createUse(dec->ownIndex(), editorFindRange(attrib, attrib));
+            }
+            else qCWarning(KDEV_PYTHON_DUCHAIN) << "No declaration created for " << attrib->attribute << "as parent is not a class";
+
+            closeInjectedContext();
+        }
     }
     else {
-        injectContext(internal.data());
-
-        Declaration* dec = visitVariableDeclaration<ClassMemberDeclaration>(
-            attrib->attribute, attrib, haveDeclaration, element.type
-        );
-        if ( dec ) {
-            dec->setRange(RangeInRevision(internal->range().start, internal->range().start));
-            dec->setAutoDeclaration(true);
-            DUChainWriteLocker lock;
-            previousContext->createUse(dec->ownIndex(), editorFindRange(attrib, attrib));
+        DUChainWriteLocker lock;
+        // the declaration is already there, just update the type
+        if ( ! attributeDeclaration->type<FunctionType>() ) {
+            auto newType = Helper::mergeTypes(attributeDeclaration->abstractType(), element.type);
+            attributeDeclaration->setAbstractType(newType);
         }
-        else qCWarning(KDEV_PYTHON_DUCHAIN) << "No declaration created for " << attrib->attribute << "as parent is not a class";
+    }
+}
 
-        closeInjectedContext();
+void DeclarationBuilder::assignToTuple(TupleAst* tuple, const DeclarationBuilder::SourceType& element) {
+    auto sources = unpackAssignmentSource(element, tuple->elements.length());
+    int ii = 0;
+    bool foundStarred = false;
+    foreach ( ExpressionAst* target, tuple->elements ) {
+        // Fallback, if we ran out of known types.
+        auto source = SourceType{
+            AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed)),
+            DeclarationPointer(),
+            false
+        };
+
+        if ( target->astType == Ast::StarredAstType ) {
+            // PEP-3132. `a, *b, c = 1, 2, 3, 4 -> b = [2, 3]`
+            // Starred expression is assigned a list of the values not used by unstarred ones.
+            DUChainReadLocker lock;
+            auto type = ExpressionVisitor::typeObjectForIntegralType<ListType>("list");
+            lock.unlock();
+            if ( !foundStarred ) { // Only allowed once, return unknown list.
+                // Count sources not used by other targets, slurp that many into list.
+                int leftovers = sources.length() - tuple->elements.length() + 1;
+                for ( ; leftovers > 0; leftovers--, ii++ ) {
+                    type->addContentType<Python::UnsureType>(sources.at(ii).type);
+                }
+            }
+            source.type = AbstractType::Ptr::staticCast(type);
+            // List is assigned to the child expression.
+            target = static_cast<StarredAst*>(target)->value;
+            foundStarred = true;
+        }
+        else if ( ii < sources.length() ) {
+            source = sources.at(ii);
+            ii++;
+        }
+        assignToUnknown(target, source);
+    }
+}
+
+void DeclarationBuilder::assignToUnknown(ExpressionAst* target, const DeclarationBuilder::SourceType& element) {
+    // Must be a nicer way to do this.
+    if ( target->astType == Ast::TupleAstType ) {
+        // Assignments of the form "a, b = 1, 2" or "a, b = c"
+        assignToTuple(static_cast<TupleAst*>(target), element);
+    }
+    else if ( target->astType == Ast::NameAstType ) {
+        // Assignments of the form "a = 3"
+        assignToName(static_cast<NameAst*>(target), element);
+    }
+    else if ( target->astType == Ast::SubscriptAstType ) {
+        // Assignments of the form "a[0] = 3"
+        assignToSubscript(static_cast<SubscriptAst*>(target), element);
+    }
+    else if ( target->astType == Ast::AttributeAstType ) {
+        // Assignments of the form "a.b = 3"
+        assignToAttribute(static_cast<AttributeAst*>(target), element);
     }
 }
 
 void DeclarationBuilder::visitAssignment(AssignmentAst* node)
 {
     AstDefaultVisitor::visitAssignment(node);
-    // Because of tuple unpacking, it is required to gather the left- and right hand side
-    // expressions / types first, then match them together in a second step.
-    const QList<ExpressionAst*>& targets = targetsOfAssignment(node->targets);
-    const QList<SourceType>& sources = sourcesOfAssignment(node->value, targets.size() > 1 ? targets.size() : -1);
 
-    // Now all the information about left- and right hand side entries is ready,
-    // and creation / updating of variables can start.
-    int i = 0;
-    foreach ( ExpressionAst* target, targets ) {
-        SourceType element(selectSource(targets, sources, i, node->value));
+    ExpressionVisitor v(currentContext());
+    v.visitNode(node->value);
+    auto sourceType = SourceType{
+        v.lastType(),
+        DeclarationPointer(Helper::resolveAliasDeclaration(v.lastDeclaration().data())),
+        v.isAlias()
+    };
 
-        // Handling the tuple unpacking stuff is done now, and we can proceed as if there was no tuple unpacking involved.
-        if ( target->astType == Ast::NameAstType ) {
-            // Assignments of the form "a = 3"
-            assignToName(static_cast<NameAst*>(target), element);
-        }
-        else if ( target->astType == Ast::SubscriptAstType ) {
-            // Assignments of the form "a[0] = 3"
-            assignToSubscript(static_cast<SubscriptAst*>(target), element);
-        }
-        else if ( target->astType == Ast::AttributeAstType ) {
-            // Assignments of the form "a.b = 3"
-            assignToAttribute(static_cast<AttributeAst*>(target), element);
-        }
-        i += 1;
+    foreach(ExpressionAst* target, node->targets) {
+        assignToUnknown(target, sourceType);
     }
 }
 
@@ -1448,6 +1447,8 @@ void DeclarationBuilder::visitClassDefinition( ClassDefinitionAst* node )
     dec->setKind(KDevelop::Declaration::Type);
     dec->clearBaseClasses();
     dec->setClassType(ClassDeclarationData::Class);
+
+    dec->setComment(getDocstring(node->body));
     
     // check whether this is a type container (list, dict, ...) or just a "normal" class
     if ( Helper::docstringContainsHint(dec, "TypeContainer") ) {
@@ -1521,8 +1522,6 @@ void DeclarationBuilder::visitClassDefinition( ClassDefinitionAst* node )
     closeContext();
     closeType();
     closeDeclaration();
-    
-    dec->setComment(getDocstring(node->body));
 }
 
 template<typename T> void DeclarationBuilder::visitDecorators(QList< Python::ExpressionAst* > decorators, T* addTo) {
@@ -1704,6 +1703,7 @@ QString DeclarationBuilder::getDocstring(QList< Python::Ast* > body) const
     {
         // If the first statement in a function/class body is a string, then that is the docstring.
         StringAst* docstring = static_cast<StringAst*>(static_cast<ExpressionAst*>(body.first())->value);
+        docstring->usedAsComment = true;
         return docstring->value.trimmed();
     }
     return QString();
@@ -1813,21 +1813,22 @@ void DeclarationBuilder::visitReturn(ReturnAst* node)
     
     if ( node->value ) {
         if ( ! hasCurrentType() ) {
-            DUChainWriteLocker lock(DUChain::lock());
+            DUChainWriteLocker lock;
             KDevelop::Problem *p = new KDevelop::Problem();
             p->setFinalLocation(DocumentRange(currentlyParsedDocument(), node->range())); // only mark first line
             p->setSource(KDevelop::IProblem::SemanticAnalysis);
             p->setDescription(i18n("Return statement not within function declaration"));
             ProblemPointer ptr(p);
             topContext()->addProblem(ptr);
-            return;
         }
-        TypePtr<FunctionType> t = currentType<FunctionType>();
-        AbstractType::Ptr encountered = v.lastType();
-        DUChainWriteLocker lock;
-        if ( t ) {
-            // Update the containing function's return type
-            t->setReturnType(Helper::mergeTypes(t->returnType(), encountered));
+        else {
+            TypePtr<FunctionType> t = currentType<FunctionType>();
+            AbstractType::Ptr encountered = v.lastType();
+            DUChainWriteLocker lock;
+            if ( t ) {
+                // Update the containing function's return type
+                t->setReturnType(Helper::mergeTypes(t->returnType(), encountered));
+            }
         }
     }
     DeclarationBuilderBase::visitReturn(node);
@@ -1851,7 +1852,7 @@ void DeclarationBuilder::visitArguments( ArgumentsAst* node )
     int firstDefaultParameterOffset = parametersCount - defaultParametersCount;
     int currentIndex = 0;
     qCDebug(KDEV_PYTHON_DUCHAIN) << "arguments:" << node->arguments.size();
-    foreach ( ArgAst* arg, node->arguments ) {
+    foreach ( ArgAst* arg, node->arguments + node->kwonlyargs ) {
         // Iterate over all the function's arguments, create declarations, and add the arguments
         // to the functions FunctionType.
         currentIndex += 1;
@@ -1891,7 +1892,7 @@ void DeclarationBuilder::visitArguments( ArgumentsAst* node )
                 argumentType = Helper::mergeTypes(paramDeclaration->abstractType(), v.lastType());
             }
         }
-        else if ( currentIndex > firstDefaultParameterOffset ) {
+        else if ( currentIndex > firstDefaultParameterOffset && currentIndex <= node->arguments.size() ) {
             // Handle arguments with default values, like def foo(bar = 3): pass
             // Find type of given default value, and assign it to the declaration
             ExpressionVisitor v(currentContext());
@@ -1959,6 +1960,20 @@ void DeclarationBuilder::visitArguments( ArgumentsAst* node )
     }
 }
 
+void DeclarationBuilder::visitString(StringAst* node) {
+    if ( node->parent && node->parent->astType == Ast::ExpressionAstType ) {
+        m_lastComment = node;
+    }
+    DeclarationBuilderBase::visitString(node);
+}
+
+void DeclarationBuilder::visitNode(Ast* node) {
+    DeclarationBuilderBase::visitNode(node);
+    if ( node && node->astType >= Ast::StatementAstType && node->astType <= Ast::LastStatementType) {
+        m_lastComment = nullptr;
+    }
+}
+
 void DeclarationBuilder::visitGlobal(GlobalAst* node)
 {
     TopDUContext* top = topContext();
