@@ -45,8 +45,8 @@ public:
     /**
      * @brief Entry function, called by KDevPlatform.
      */
-    virtual ReferencedTopDUContext build(const IndexedString& url, Ast* node,
-                                         ReferencedTopDUContext updateContext = ReferencedTopDUContext());
+    ReferencedTopDUContext build(const IndexedString& url, Ast* node,
+                                 ReferencedTopDUContext updateContext = ReferencedTopDUContext()) override;
 
     /**
      * @brief Set whether the current running pass is the first or the second one.
@@ -83,34 +83,43 @@ public:
 
 protected:
     /// AST visitor functions
-    virtual void visitClassDefinition(ClassDefinitionAst* node);
-    virtual void visitFunctionDefinition(FunctionDefinitionAst* node);
-    virtual void visitAssignment(AssignmentAst* node);
-    virtual void visitFor(ForAst* node);
-    virtual void visitImport(ImportAst* node);
-    virtual void visitImportFrom(ImportFromAst* node);
-    virtual void visitArguments(ArgumentsAst* node);
-    virtual void visitExceptionHandler(ExceptionHandlerAst* node);
-    virtual void visitReturn(ReturnAst* node);
-    virtual void visitCode(CodeAst* node);
-    virtual void visitCall(CallAst* node);
-    virtual void visitYield(YieldAst* node);
-    virtual void visitWithItem(WithItemAst* node);
-    virtual void visitLambda(LambdaAst* node);
-    virtual void visitComprehension(ComprehensionAst* node);
-    virtual void visitGlobal(GlobalAst* node);
-    virtual void visitAssertion(AssertionAst* node);
-    virtual void visitIf(IfAst* node);
+    void visitClassDefinition(ClassDefinitionAst* node) override;
+    void visitFunctionDefinition(FunctionDefinitionAst* node) override;
+    void visitAssignment(AssignmentAst* node) override;
+    void visitFor(ForAst* node) override;
+    void visitImport(ImportAst* node) override;
+    void visitImportFrom(ImportFromAst* node) override;
+    void visitArguments(ArgumentsAst* node) override;
+    void visitExceptionHandler(ExceptionHandlerAst* node) override;
+    void visitReturn(ReturnAst* node) override;
+    void visitCode(CodeAst* node) override;
+    void visitCall(CallAst* node) override;
+    void visitYield(YieldAst* node) override;
+    void visitWithItem(WithItemAst* node) override;
+    void visitLambda(LambdaAst* node) override;
+    void visitComprehension(ComprehensionAst* node) override;
+    void visitGlobal(GlobalAst* node) override;
+    void visitAssertion(AssertionAst* node) override;
+    void visitIf(IfAst* node) override;
+    void visitString(StringAst* node) override;
+    void visitNode(Ast* node) override;
 
 protected:
+    enum VisitVariableFlags {
+        NoVisitVariableFlags = 0x0,
+        AbortIfReopenMismatch = 0x1
+    };
     /// Visitor helper functions
     template<typename T> void visitDecorators(QList<ExpressionAst*> decorators, T* addTo);
-    template<typename T> T* visitVariableDeclaration(Python::Ast* node, Declaration* previous = 0,
-                                                     AbstractType::Ptr type = AbstractType::Ptr(0));
-    template<typename T> T* visitVariableDeclaration(Identifier* node, Ast* originalAst = 0, Declaration* previous = 0,
-                                                     AbstractType::Ptr type = AbstractType::Ptr(0));
+    template<typename T> T* visitVariableDeclaration(Python::Ast* node, Declaration* previous=nullptr,
+                                                     AbstractType::Ptr type = AbstractType::Ptr(),
+                                                     VisitVariableFlags flags=NoVisitVariableFlags);
+    template<typename T> T* visitVariableDeclaration(Identifier* node, Ast* originalAst = 0, Declaration* previous=nullptr,
+                                                     AbstractType::Ptr type = AbstractType::Ptr(),
+                                                     VisitVariableFlags flags=NoVisitVariableFlags);
     template<typename T> T* visitVariableDeclaration(Identifier* node, RangeInRevision range,
-                                                     AbstractType::Ptr type = AbstractType::Ptr(0));
+                                                     AbstractType::Ptr type = AbstractType::Ptr(),
+                                                     VisitVariableFlags flags=NoVisitVariableFlags);
 
 protected:
     /**
@@ -160,7 +169,7 @@ protected:
     /**
      * @brief Attempts to determine the unpacked type(s) of the right-hand side of an assignment.
      *
-     * @param items The expression on the right-hand side of the assignment
+     * @param source The source type on the right-hand side of the assignment
      * @param fillWhenLengthMissing If non-zero and the object in @p items has an indefinite length but
      *                              a definite content type, return this many copies of that type
      *
@@ -168,20 +177,7 @@ protected:
      *          a, b = "Foo", 3.5 (@p fillWhenLengthMissing ignored here because the right-hand side
      *                  has a definite amount of items) -> returns str, float
      */
-    QList<SourceType> sourcesOfAssignment(ExpressionAst* items, int fillWhenLengthMissing=-1) const;
-
-    /**
-     * @brief Given a list of targets (@see targetsOfAssignment) and sources (@see sourcesOfAssignment) and a
-     *        @p index, matches the sources with the targets and returns the type which belongs to @p index.
-     *
-     * @param targets Left-hand side of the assignment
-     * @param sources Right-hand side of the assignment
-     * @param index index of the item to return the type for
-     * @param rhs Pass the right-hand side of the assignment again so that its compound type can be used in case
-     *            it consists of more than one element and unpacking is not possible; example: a = 1, 2
-     */
-    SourceType selectSource(const QList<ExpressionAst*>& targets, const QList<SourceType>& sources,
-                            int index, ExpressionAst* rhs) const;
+    QVector<SourceType> unpackAssignmentSource(const SourceType& source, int fillWhenLengthMissing) const;
 
     /**
       * @brief Handle a variable assignment to @p name and give it the type @p element.
@@ -198,6 +194,16 @@ protected:
      */
     void assignToAttribute(AttributeAst* attribute, const SourceType& element);
 
+    /**
+     * @brief Handle assignment to a target @p target with rhs type @p element.
+     */
+    void assignToTuple(TupleAst* tuple, const SourceType& element);
+
+    /**
+     * @brief Handle assignment to a target @p target with rhs type @p element.
+     */
+    void assignToUnknown(ExpressionAst* target, const SourceType& element);
+
     /**
      * @brief Find all existing declarations for the identifier @p node
      */
@@ -285,7 +291,7 @@ private:
         decl->setAlwaysForceDirect(true);
         return decl;
     };
-    virtual void closeDeclaration();
+    void closeDeclaration() override;
 
 private:
     /// HACK: List of items to delete after parsing finishes, to work around the built-in cleanup logic
@@ -295,6 +301,8 @@ private:
     StructureType::Ptr m_currentClassType;
     // missing modules, for not reporting them as unknown variables
     QVector<IndexedString> m_missingModules;
+
+    StringAst* m_lastComment = nullptr;
 };
 
 }
