@@ -21,6 +21,7 @@
 
 #include <QList>
 #include <QProcess>
+#include <QSettings>
 #include <QStandardPaths>
 
 #include <QDebug>
@@ -37,6 +38,7 @@
 #include <language/duchain/types/typeutils.h>
 #include <language/backgroundparser/backgroundparser.h>
 #include <interfaces/iproject.h>
+#include <interfaces/iprojectcontroller.h>
 #include <interfaces/icore.h>
 #include <interfaces/ilanguagecontroller.h>
 #include <interfaces/idocumentcontroller.h>
@@ -58,7 +60,7 @@ using namespace KDevelop;
 
 namespace Python {
 
-QList<QUrl> Helper::cachedCustomIncludes;
+QMap<IProject*, QList<QUrl>> Helper::cachedCustomIncludes;
 QList<QUrl> Helper::cachedSearchPaths;
 QList<QUrl> Helper::projectSearchPaths;
 QStringList Helper::dataDirs;
@@ -101,7 +103,7 @@ IndexedDeclaration Helper::declarationUnderCursor(bool allowUse)
         DUChainReadLocker lock;
         const auto cursor = view->cursorPosition();
         if ( allowUse ) {
-            return DUChainUtils::itemUnderCursor(doc->url(), cursor);
+            return IndexedDeclaration(DUChainUtils::itemUnderCursor(doc->url(), cursor).declaration);
         }
         else {
             return DUChainUtils::declarationInLine(cursor, DUChainUtils::standardContextForUrl(doc->url()));
@@ -126,10 +128,15 @@ Declaration* Helper::accessAttribute(Declaration* accessed, const QString& attri
     for ( auto type: structureTypes ) {
         QList<DUContext*> searchContexts = Helper::internalContextsForClass(type, current->topContext());
         for ( DUContext* c: searchContexts ) {
-            QList< Declaration* > found = c->findDeclarations(KDevelop::Identifier(attribute),
-                                                              CursorInRevision::invalid(),
-                                                              current->topContext(), DUContext::DontSearchInParent);
-            if ( ! found.isEmpty() ) {
+            auto found = c->findDeclarations(KDevelop::Identifier(attribute),
+                                             CursorInRevision::invalid(),
+                                             current->topContext(), DUContext::DontSearchInParent);
+            std::reverse(found.begin(), found.end());
+            // never consider decls from the builtins
+            if ( ! found.isEmpty() && (
+                   found.first()->topContext() != Helper::getDocumentationFileContext() ||
+                   c->topContext() == Helper::getDocumentationFileContext() ) )
+            {
                 return found.first();
             }
         }
@@ -142,79 +149,12 @@ AbstractType::Ptr Helper::resolveAliasType(const AbstractType::Ptr eventualAlias
     return TypeUtils::resolveAliasType(eventualAlias);
 }
 
-AbstractType::Ptr Helper::extractTypeHints(AbstractType::Ptr type, TopDUContext* current)
+AbstractType::Ptr Helper::extractTypeHints(AbstractType::Ptr type)
 {
-    UnsureType::Ptr result(new UnsureType());
-    unsigned short maxHints = 7;
-    if ( HintedType::Ptr hinted = type.cast<HintedType>() ) {
-        if ( hinted->isValid(current) && isUsefulType(hinted.cast<AbstractType>()) ) {
-            result->addType(type->indexed());
-        }
-    }
-    else if ( UnsureType::Ptr unsure = type.cast<UnsureType>() ) {
-        int len = unsure->typesSize();
-        for ( int i = 0; i < len && i < maxHints; i++ ) {
-            if ( HintedType::Ptr hinted = unsure->types()[i].abstractType().cast<HintedType>() ) {
-                if ( hinted->isValid(current) ) {
-                    qCDebug(KDEV_PYTHON_DUCHAIN) << "Adding type hint (multi): " << hinted->toString();
-                    result->addType(hinted->indexed());
-                }
-                else {
-                    qCDebug(KDEV_PYTHON_DUCHAIN) << "Discarding type hint (multi): " << hinted->toString();
-                    maxHints += 1;
-                }
-            }
-            else {
-                maxHints += 1;
-            }
-        }
-    }
-    else if ( IndexedContainer::Ptr indexed = type.cast<IndexedContainer>() ) {
-        // TODO this is bad because it is slow. Make it faster!
-        // TODO this would really be an important thing, as it will be called quite often.
-        // TODO "how" is simple, just avoid cloning stuff if nothing needs to be changed (i.e. no hints exist).
-        IndexedContainer::Ptr edit = IndexedContainer::Ptr(static_cast<IndexedContainer*>(indexed->clone()));
-        for ( int i = 0; i < indexed->typesCount(); i++ ) {
-            if ( HintedType::Ptr p = indexed->typeAt(i).abstractType().cast<HintedType>() ) {
-                if ( ! p->isValid(current) ) {
-                    edit->replaceType(i, AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed)));
-                }
-            }
-        }
-        return edit.cast<AbstractType>();
-    }
-    else if ( auto variable = type.cast<ListType>() ) {
-        auto edit = ListType::Ptr(static_cast<ListType*>(variable->clone()));
-        UnsureType::Ptr newContentType(new UnsureType());
-        AbstractType::Ptr oldContentType = edit->contentType().abstractType();
-        bool isHint = false;
-        if ( oldContentType ) {
-            if ( UnsureType::Ptr oldUnsure = oldContentType.cast<UnsureType>() ) {
-                for ( unsigned int i = 0; i < oldUnsure->typesSize(); i++ ) {
-                    if ( HintedType::Ptr hinted = oldUnsure->types()[i].abstractType().cast<HintedType>() ) {
-                        isHint = true;
-                        if ( ! hinted->isValid(current) ) {
-                            continue;
-                        }
-                    }
-                    newContentType->addType(oldUnsure->types()[i]);
-                }
-            }
-            else if ( HintedType::Ptr hinted = oldContentType.cast<HintedType>() ) {
-                isHint = true;
-                if ( hinted->isValid(current) ) {
-                    newContentType->addType(hinted->indexed());
-                }
-            }
-            if ( ! isHint ) {
-                return result.cast<AbstractType>();
-            }
-            edit->replaceContentType(newContentType.cast<AbstractType>());
-//             edit->replaceKeyType(variable->keyType().abstractType()); // TODO re-enable?
-        }
-        return edit.cast<AbstractType>();
-    }
-    return result.cast<AbstractType>();
+    return Helper::foldTypes(Helper::filterType<AbstractType>(type, [](AbstractType::Ptr t) -> bool {
+        auto hint = t.cast<HintedType>();
+        return !hint || hint->isValid();
+    }));
 }
 
 Helper::FuncInfo Helper::functionDeclarationForCalledDeclaration(DeclarationPointer ptr)
@@ -277,7 +217,7 @@ Declaration* Helper::declarationForName(const QualifiedIdentifier& identifier, c
         do {
             declaration = importedLocalDeclarations.last();
             importedLocalDeclarations.pop_back();
-            if ( !declaration || declaration->context()->type() == DUContext::Class ) {
+            if ( !declaration || (declaration->context()->type() == DUContext::Class && context->type() != DUContext::Function) ) {
                 declaration = 0;
             }
             if ( importedLocalDeclarations.isEmpty() ) {
@@ -331,7 +271,7 @@ Declaration* Helper::resolveAliasDeclaration(Declaration* decl)
 
 QStringList Helper::getDataDirs() {
     if ( Helper::dataDirs.isEmpty() ) {
-        Helper::dataDirs = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, "kdevpythonsupport/documentation_files",QStandardPaths::LocateDirectory);
+        Helper::dataDirs = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, "kdevpythonsupport/documentation_files", QStandardPaths::LocateDirectory);
     }
     return Helper::dataDirs;
 }
@@ -410,10 +350,8 @@ QUrl Helper::getCorrectionFile(const QUrl& document)
             auto doc = document.path();
             auto relative = _relativePath(base, doc);
             auto fullPath = correctionFileDir + "/" + relative;
-            // TODO QUrl: cleanPath?
-
             if ( QFile::exists(fullPath) ) {
-                return QUrl::fromLocalFile(fullPath);
+                return QUrl::fromLocalFile(fullPath).adjusted(QUrl::NormalizePathSegments);
             }
         }
     }
@@ -427,7 +365,7 @@ QUrl Helper::getLocalCorrectionFile(const QUrl& document)
     }
 
     auto absolutePath = QUrl();
-    foreach ( const auto& basePath, Helper::getSearchPaths({}) ) {
+    foreach ( const auto& basePath, Helper::getSearchPaths(QUrl()) ) {
         if ( ! basePath.isParentOf(document) ) {
             continue;
         }
@@ -437,7 +375,55 @@ QUrl Helper::getLocalCorrectionFile(const QUrl& document)
     }
     return absolutePath;
 }
-    
+
+QString getPythonExecutablePath() {
+    // Find python 3 (https://www.python.org/dev/peps/pep-0394/)
+    auto result = QStandardPaths::findExecutable("python" PYTHON_VERSION_MAJOR "." PYTHON_VERSION_MINOR);
+    if ( ! result.isEmpty() ) {
+        return result;
+    }
+    result = QStandardPaths::findExecutable("python" PYTHON_VERSION_MAJOR);
+    if ( ! result.isEmpty() ) {
+        return result;
+    }
+    result = QStandardPaths::findExecutable("python");
+    if ( ! result.isEmpty() ) {
+        return result;
+    }
+
+#ifdef Q_OS_WIN
+    QStringList extraPaths;
+    // Check for default CPython installation path, because the
+    // installer does not add the path to $PATH.
+    QStringList keys = {
+        "HKEY_LOCAL_MACHINE\\Software\\Python\\PythonCore\\PYTHON_VERSION\\InstallPath",
+        "HKEY_LOCAL_MACHINE\\Software\\Python\\PythonCore\\PYTHON_VERSION-32\\InstallPath",
+        "HKEY_CURRENT_USER\\Software\\Python\\PythonCore\\PYTHON_VERSION\\InstallPath",
+        "HKEY_CURRENT_USER\\Software\\Python\\PythonCore\\PYTHON_VERSION-32\\InstallPath"
+    };
+    auto version = QString(PYTHON_VERSION_MAJOR) + "." + PYTHON_VERSION_MINOR;
+    foreach ( QString key, keys ) {
+        key.replace("PYTHON_VERSION", version);
+        QSettings base(key.left(key.indexOf("Python")), QSettings::NativeFormat);
+        if ( ! base.childGroups().contains("Python") ) {
+            continue;
+        }
+        QSettings keySettings(key, QSettings::NativeFormat);
+        auto path = keySettings.value("Default").toString();
+        if ( ! path.isEmpty() ) {
+            extraPaths << path;
+            break;
+        }
+    }
+    result = QStandardPaths::findExecutable("python", extraPaths);
+    if ( ! result.isEmpty() ) {
+        return result;
+    }
+#endif
+    // fallback
+    return PYTHON_EXECUTABLE;
+}
+
 QList<QUrl> Helper::getSearchPaths(const QUrl& workingOnDocument)
 {
     QMutexLocker lock(&Helper::cacheMutex);
@@ -446,8 +432,10 @@ QList<QUrl> Helper::getSearchPaths(const QUrl& workingOnDocument)
     // and also add custom include paths that are defined in the projects
 
     {
+        auto project = ICore::self()->projectController()->findProjectForUrl(workingOnDocument);
         QMutexLocker lock(&Helper::projectPathLock);
         searchPaths << Helper::projectSearchPaths;
+        searchPaths << Helper::cachedCustomIncludes.value(project);
     }
     
     foreach ( const QString& path, getDataDirs() ) {
@@ -457,13 +445,13 @@ QList<QUrl> Helper::getSearchPaths(const QUrl& workingOnDocument)
     if ( cachedSearchPaths.isEmpty() ) {
         qCDebug(KDEV_PYTHON_DUCHAIN) << "*** Gathering search paths...";
         QStringList getpath;
-        getpath << "-c" << "import sys; sys.stdout.write(':'.join(sys.path))";
+        getpath << "-c" << "import sys; sys.stdout.write('$|$'.join(sys.path))";
         
         QProcess python;
-        python.start(QLatin1String(PYTHON_EXECUTABLE), getpath);
+        python.start(getPythonExecutablePath(), getpath);
         python.waitForFinished(1000);
-        QByteArray pythonpath = python.readAllStandardOutput();
-        QList<QByteArray> paths = pythonpath.split(':');
+        QString pythonpath = QString::fromUtf8(python.readAllStandardOutput());
+        auto paths = pythonpath.split("$|$");
         paths.removeAll("");
         
         if ( ! pythonpath.isEmpty() ) {
